@@ -1,11 +1,8 @@
 package Data::MessagePack::PP;
-
-use 5.008000;
+use 5.008001;
 use strict;
 use Carp ();
 
-our $VERSION = '0.15';
-
 # See also
 # http://redmine.msgpack.org/projects/msgpack/wiki/FormatSpec
 # http://cpansearch.perl.org/src/YAPPO/Data-Model-0.00006/lib/Data/Model/Driver/Memcached.pm
@@ -15,37 +12,47 @@ our $VERSION = '0.15';
 package
     Data::MessagePack;
 
-use Scalar::Util qw( blessed );
 use strict;
 use B ();
 
 BEGIN {
+    my $unpack_int64_slow;
+    my $unpack_uint64_slow;
+
+    if(!eval { pack 'Q', 1 }) { # don't have quad types
+        $unpack_int64_slow = sub {
+            require Math::BigInt;
+            my $high = unpack_uint32( $_[0], $_[1] );
+            my $low  = unpack_uint32( $_[0], $_[1] + 4);
+
+            if($high < 0xF0000000) { # positive
+                $high = Math::BigInt->new( $high );
+                $low  = Math::BigInt->new( $low  );
+                return +($high << 32 | $low)->bstr;
+            }
+            else { # negative
+                $high = Math::BigInt->new( ~$high );
+                $low  = Math::BigInt->new( ~$low  );
+                return +( -($high << 32 | $low + 1) )->bstr;
+            }
+        };
+        $unpack_uint64_slow = sub {
+            require Math::BigInt;
+            my $high = Math::BigInt->new( unpack_uint32(  $_[0], $_[1]) );
+            my $low  = Math::BigInt->new( unpack_uint32( $_[0], $_[1] + 4) );
+            return +($high << 32 | $low)->bstr;
+        };
+    }
+
+    *unpack_uint16 = sub { return unpack 'n', substr( $_[0], $_[1], 2 ) };
+    *unpack_uint32 = sub { return unpack 'N', substr( $_[0], $_[1], 4 ) };
+
     # for pack and unpack compatibility
     if ( $] < 5.010 ) {
         # require $Config{byteorder}; my $bo_is_le = ( $Config{byteorder} =~ /^1234/ );
         # which better?
         my $bo_is_le = unpack ( 'd', "\x00\x00\x00\x00\x00\x00\xf0\x3f") == 1; # 1.0LE
-        # In really, since 5.9.2 '>' is introduced. but 'n!' and 'N!'?
-        *pack_uint64 = $bo_is_le ? sub {
-            my @v = unpack( 'V2', pack( 'Q', $_[0] ) );
-            return pack 'CN2', 0xcf, @v[1,0];
-        } : sub { pack 'CQ', 0xcf, $_[0]; };
-        *pack_int64 = $bo_is_le ? sub {
-            my @v = unpack( 'V2', pack( 'q', $_[0] ) );
-            return pack 'CN2', 0xd3, @v[1,0];
-        } : sub { pack 'Cq', 0xd3, $_[0]; };
-        *pack_double = $bo_is_le ? sub {
-            my @v = unpack( 'V2', pack( 'd', $_[0] ) );
-            return pack 'CN2', 0xcb, @v[1,0];
-        } : sub { pack 'Cd', 0xcb, $_[0]; };
-        *unpack_float = $bo_is_le ? sub {
-            my @v = unpack( 'v2', substr( $_[0], $_[1], 4 ) );
-            return unpack( 'f', pack( 'n2', @v[1,0] ) );
-        } : sub { return unpack( 'f', substr( $_[0], $_[1], 4 ) ); };
-        *unpack_double = $bo_is_le ? sub {
-            my @v = unpack( 'V2', substr( $_[0], $_[1], 8 ) );
-            return unpack( 'd', pack( 'N2', @v[1,0] ) );
-        } : sub { return unpack( 'd', substr( $_[0], $_[1], 8 ) ); };
+
         *unpack_int16  = sub {
             my $v = unpack 'n', substr( $_[0], $_[1], 2 );
             return $v ? $v - 0x10000 : 0;
@@ -55,33 +62,69 @@ BEGIN {
             my $v = unpack 'N', substr( $_[0], $_[1], 4 );
             return $v ? $v - 0x100000000 : 0;
         };
-        *unpack_int64 = $bo_is_le ? sub {
-            my @v = unpack( 'V*', substr( $_[0], $_[1], 8 ) );
-            return unpack( 'q', pack( 'N2', @v[1,0] ) );
-        } : sub { pack 'q', substr( $_[0], $_[1], 8 ); };
-        *unpack_uint64 = $bo_is_le ? sub {
-            my @v = unpack( 'V*', substr( $_[0], $_[1], 8 ) );
-            return unpack( 'Q', pack( 'N2', @v[1,0] ) );
-        } : sub { pack 'Q', substr( $_[0], $_[1], 8 ); };
+
+        # In reality, since 5.9.2 '>' is introduced. but 'n!' and 'N!'?
+        if($bo_is_le) {
+            *pack_uint64 = sub {
+                my @v = unpack( 'V2', pack( 'Q', $_[0] ) );
+                return pack 'CN2', 0xcf, @v[1,0];
+            };
+            *pack_int64 = sub {
+                my @v = unpack( 'V2', pack( 'q', $_[0] ) );
+                return pack 'CN2', 0xd3, @v[1,0];
+            };
+            *pack_double = sub {
+                my @v = unpack( 'V2', pack( 'd', $_[0] ) );
+                return pack 'CN2', 0xcb, @v[1,0];
+            };
+
+            *unpack_float = sub {
+                my @v = unpack( 'v2', substr( $_[0], $_[1], 4 ) );
+                return unpack( 'f', pack( 'n2', @v[1,0] ) );
+            };
+            *unpack_double = sub {
+                my @v = unpack( 'V2', substr( $_[0], $_[1], 8 ) );
+                return unpack( 'd', pack( 'N2', @v[1,0] ) );
+            };
+
+            *unpack_int64 = $unpack_int64_slow ||_sub {
+                my @v = unpack( 'V*', substr( $_[0], $_[1], 8 ) );
+                return unpack( 'q', pack( 'N2', @v[1,0] ) );
+            };
+            *unpack_uint64 = $unpack_uint64_slow || sub {
+                my @v = unpack( 'V*', substr( $_[0], $_[1], 8 ) );
+                return unpack( 'Q', pack( 'N2', @v[1,0] ) );
+            };
+        }
+        else { # big endian
+            *pack_uint64   = sub { return pack 'CQ', 0xcf, $_[0]; };
+            *pack_int64    = sub { return pack 'Cq', 0xd3, $_[0]; };
+            *pack_double   = sub { return pack 'Cd', 0xcb, $_[0]; };
+
+            *unpack_float  = sub { return unpack( 'f', substr( $_[0], $_[1], 4 ) ); };
+            *unpack_double = sub { return unpack( 'd', substr( $_[0], $_[1], 8 ) ); };
+            *unpack_int64  = $unpack_int64_slow  || sub { pack 'q', substr( $_[0], $_[1], 8 ); };
+            *unpack_uint64 = $unpack_uint64_slow || sub { pack 'Q', substr( $_[0], $_[1], 8 ); };
+        }
     }
     else {
         *pack_uint64   = sub { return pack 'CQ>', 0xcf, $_[0]; };
         *pack_int64    = sub { return pack 'Cq>', 0xd3, $_[0]; };
         *pack_double   = sub { return pack 'Cd>', 0xcb, $_[0]; };
+
         *unpack_float  = sub { return unpack( 'f>', substr( $_[0], $_[1], 4 ) ); };
         *unpack_double = sub { return unpack( 'd>', substr( $_[0], $_[1], 8 ) ); };
         *unpack_int16  = sub { return unpack( 'n!', substr( $_[0], $_[1], 2 ) ); };
         *unpack_int32  = sub { return unpack( 'N!', substr( $_[0], $_[1], 4 ) ); };
-        *unpack_int64  = sub { return unpack( 'q>', substr( $_[0], $_[1], 8 ) ); };
-        *unpack_uint64 = sub { return unpack( 'Q>', substr( $_[0], $_[1], 8 ) ); };
-    }
-    # for 5.8 etc.
-    unless ( defined &utf8::is_utf8 ) {
-       require Encode;
-       *utf8::is_utf8 = *Encode::is_utf8;
+
+        *unpack_int64  = $unpack_int64_slow  || sub { return unpack( 'q>', substr( $_[0], $_[1], 8 ) ); };
+        *unpack_uint64 = $unpack_uint64_slow || sub { return unpack( 'Q>', substr( $_[0], $_[1], 8 ) ); };
     }
 }
 
+sub _unexpected {
+    Carp::confess("Unexpected " . sprintf(shift, @_) . " found");
+}
 
 #
 # PACK
@@ -90,11 +133,11 @@ BEGIN {
 {
     no warnings 'recursion';
 
-    my $max_depth;
+    our $_max_depth;
 
-sub pack {
+sub pack :method {
     Carp::croak('Usage: Data::MessagePack->pack($dat [,$max_depth])') if @_ < 2;
-    $max_depth = defined $_[2] ? $_[2] : 512; # init
+    $_max_depth = defined $_[2] ? $_[2] : 512; # init
     return _pack( $_[1] );
 }
 
@@ -102,42 +145,42 @@ sub pack {
 sub _pack {
     my ( $value ) = @_;
 
+    local $_max_depth = $_max_depth - 1;
+
+    if ( $_max_depth < 0 ) {
+        Carp::croak("perl structure exceeds maximum nesting level (max_depth set too low?)");
+    }
+
     return CORE::pack( 'C', 0xc0 ) if ( not defined $value );
 
-    my $b_obj = B::svref_2object( ref $value ? $value : \$value );
-
-    if ( $b_obj->isa('B::AV') ) {
+    if ( ref($value) eq 'ARRAY' ) {
         my $num = @$value;
         my $header = 
               $num < 16          ? CORE::pack( 'C',  0x90 + $num )
             : $num < 2 ** 16 - 1 ? CORE::pack( 'Cn', 0xdc,  $num )
             : $num < 2 ** 32 - 1 ? CORE::pack( 'CN', 0xdd,  $num )
-            : die "" # don't arrivie here
+            : _unexpected("number %d", $num)
         ;
-        if ( --$max_depth <= 0 ) {
-            Carp::croak("perl structure exceeds maximum nesting level (max_depth set too low?)");
-        }
         return join( '', $header, map { _pack( $_ ) } @$value );
     }
 
-    elsif ( $b_obj->isa('B::HV') ) {
+    elsif ( ref($value) eq 'HASH' ) {
         my $num = keys %$value;
         my $header = 
               $num < 16          ? CORE::pack( 'C',  0x80 + $num )
             : $num < 2 ** 16 - 1 ? CORE::pack( 'Cn', 0xde,  $num )
             : $num < 2 ** 32 - 1 ? CORE::pack( 'CN', 0xdf,  $num )
-            : die "" # don't arrivie here
+            : _unexpected("number %d", $num)
         ;
-        if ( --$max_depth <= 0 ) {
-            Carp::croak("perl structure exceeds maximum nesting level (max_depth set too low?)");
-        }
         return join( '', $header, map { _pack( $_ ) } %$value );
     }
 
-    elsif ( blessed( $value ) and blessed( $value ) eq 'Data::MessagePack::Boolean' ) {
-        return  CORE::pack( 'C', $$value ? 0xc3 : 0xc2 );
+    elsif ( ref( $value ) eq 'Data::MessagePack::Boolean' ) {
+        return  CORE::pack( 'C', ${$value} ? 0xc3 : 0xc2 );
     }
 
+
+    my $b_obj = B::svref_2object( \$value );
     my $flags = $b_obj->FLAGS;
 
     if ( $flags & ( B::SVf_IOK | B::SVp_IOK ) ) {
@@ -158,7 +201,6 @@ sub _pack {
         }
 
     }
-
     elsif ( $flags & B::SVf_POK ) { # raw / check needs before dboule
 
         if ( $Data::MessagePack::PreferInteger ) {
@@ -181,19 +223,17 @@ sub _pack {
               $num < 32          ? CORE::pack( 'C',  0xa0 + $num )
             : $num < 2 ** 16 - 1 ? CORE::pack( 'Cn', 0xda, $num )
             : $num < 2 ** 32 - 1 ? CORE::pack( 'CN', 0xdb, $num )
-            : die "" # don't arrivie here
+            : _unexpected_number($num)
         ;
 
         return $header . $value;
 
     }
-
     elsif ( $flags & ( B::SVf_NOK | B::SVp_NOK ) ) { # double only
         return pack_double( $value );
     }
-
     else {
-        die "???";
+        _unexpected("data type %s", $b_obj);
     }
 
 }
@@ -209,7 +249,7 @@ sub _pack {
 
     my $p; # position variables for speed.
 
-sub unpack {
+sub unpack :method {
     $p = 0; # init
     _unpack( $_[1] );
 }
@@ -270,11 +310,11 @@ sub _unpack {
     }
     elsif ( $byte == 0xcd ) { # uint16
         $p += 2;
-        return CORE::unpack 'n', substr( $value, $p - 2, 2 );
+        return unpack_uint16( $value, $p - 2 );
     }
     elsif ( $byte == 0xce ) { # unit32
         $p += 4;
-        return CORE::unpack 'N', substr( $value, $p - 4, 4 );
+        return unpack_uint32( $value, $p - 4 );
     }
     elsif ( $byte == 0xcf ) { # unit64
         $p += 8;
@@ -337,7 +377,7 @@ sub _unpack {
     }
 
     else {
-        die "???";
+        _unexpected("byte 0x%02x", $byte);
     }
 
 }
@@ -356,7 +396,7 @@ package
 use strict;
 
 sub new {
-    bless { stack => [] }, shift;
+    bless { pos => 0 }, shift;
 }
 
 
@@ -370,25 +410,30 @@ sub execute_limit {
 
 sub execute {
     my ( $self, $data, $offset, $limit ) = @_;
+    $offset ||= 0;
     my $value = substr( $data, $offset, $limit ? $limit : length $data );
     my $len   = length $value;
 
+    $self->{data} .= $value;
+    local $self->{stack} = [];
+
     $p = 0;
 
-    while ( $len > $p ) {
-        _count( $self, $value ) or last;
+    LOOP: while ( length($self->{data}) > $p ) {
+        _count( $self, $self->{data} ) or last;
 
-        if ( @{ $self->{stack} } > 0 ) {
-            pop @{ $self->{stack} } if --$self->{stack}->[-1] == 0;
+        while ( @{ $self->{stack} } > 0 && --$self->{stack}->[-1] == 0) {
+            pop @{ $self->{stack} };
+        }
+
+        if (@{$self->{stack}} == 0) {
+            $self->{is_finished}++;
+            last LOOP;
         }
     }
+    $self->{pos} = $p;
 
-    if ( $len == $p ) {
-        $self->{ data } .= substr( $value, 0, $p );
-        $self->{ remain } = undef;
-    }
-
-    return $p;
+    return $p + $offset;
 }
 
 
@@ -410,7 +455,9 @@ sub _count {
             $num = $byte & ~0x90;
         }
 
-        push @{ $self->{stack} }, $num + 1;
+        if (defined($num) && $num > 0) {
+            push @{ $self->{stack} }, $num + 1;
+        }
 
         return 1;
     }
@@ -429,7 +476,9 @@ sub _count {
             $num = $byte & ~0x80;
         }
 
-        push @{ $self->{stack} }, $num * 2 + 1; # a pair
+        if ($num > 0) {
+            push @{ $self->{stack} }, $num * 2 + 1; # a pair
+        }
 
         return 1;
     }
@@ -447,7 +496,7 @@ sub _count {
             : $byte == 0xcd ? 2
             : $byte == 0xce ? 4
             : $byte == 0xcf ? 8
-            : die;
+            : _unexpected("byte 0x%02x", $byte);
         return 1;
     }
 
@@ -456,7 +505,7 @@ sub _count {
             : $byte == 0xd1 ? 2
             : $byte == 0xd2 ? 4
             : $byte == 0xd3 ? 8
-            : die;
+            : _unexpected("byte 0x%02x", $byte);
         return 1;
     }
 
@@ -487,7 +536,7 @@ sub _count {
     }
 
     else {
-        die "???";
+        _unexpected("byte 0x%02x", $byte);
     }
 
     return 0;
@@ -497,22 +546,19 @@ sub _count {
 
 
 sub data {
-    my $data = Data::MessagePack->unpack( $_[0]->{ data } );
-    $_[0]->reset;
-    return $data;
+    return Data::MessagePack->unpack( substr($_[0]->{ data }, 0, $_[0]->{pos}) );
 }
 
 
 sub is_finished {
     my ( $self ) = @_;
-    ( scalar( @{ $self->{stack} } ) or defined $self->{ remain } ) ? 0 : 1;
+    return $self->{is_finished};
 }
 
-
-sub reset {
-    $_[0]->{ stack }  = [];
+sub reset :method {
     $_[0]->{ data }   = undef;
-    $_[0]->{ remain } = undef;
+    $_[0]->{ pos }    = 0;
+    $_[0]->{ is_finished } = 0;
 }
 
 1;
